<?php
namespace Controllers\Auth;

use Core\Auth;
use Core\View;
use Core\BaseController;
use Core\Lang;
use Core\Logger;
use Mails\WelcomeEmail;
use Models\User;

class RegisterController extends BaseController {

    public function __construct() {
        // Seul les utilisateurs non authentifiés peuvent accéder à toutes ces pages
        $this->guestOnly();
    }

    public function index() {

        $step = $this->request()->get('step') ?? 1;
        if((int)$step > 1) {
            if(!isset($_SESSION['register']['step1'])) {
                $this->request()->flash('error', Lang::get('auth_register_error_step1'));
                return $this->redirect(route('register.index', ['step' => 1]));
            }
        }

        return View::render('auth.register_step' . $step, [
            'step' => $step
        ]);
    }

    public function register() {

        $step = $this->request()->post('step') ?? $this->request()->get('step') ?? 1;

        switch((int)$step) {
            // Questions
            case 1:
                $inputs = $this->request()->all()->only([
                    'date_naissance',
                    'nom',
                    'prenom',
                    'prenom_etat_civil',
                    'premier_chien',
                    'premier_amour',
                    'plat_prefere',
                    'film_prefere',
                    'frere_jumeau',
                ]);

                $errors = $this->request()->validate($inputs, [
                    'date_naissance' => 'required|date',
                    'nom' => 'required|string',
                    'prenom' => 'required|string',
                    'prenom_etat_civil' => 'required|string',
                    'premier_chien' => 'required|string',
                    'premier_amour' => 'required|string',
                    'plat_prefere' => 'required|string',
                    'film_prefere' => 'required|string',
                    'frere_jumeau' => 'required|string',
                ]);

                if(count($errors) > 0) {
                    return $this->redirect(route('register.index', ['step' => 1]));
                }

                $valid_values = [
                    'date_naissance' => '1996-07-04',
                    'nom' => 'bucaille',
                    'prenom' => 'thommy',
                    'prenom_etat_civil' => 'tommy',
                    'premier_chien' => 'pepette',
                    'premier_amour' => 'jennifer',
                    'plat_prefere' => 'tartiflette',
                    'film_prefere' => 'stargate',
                    'frere_jumeau' => 'johnny',
                ];

                foreach($inputs as $key => $value) {
                    $value = strtolower($value);
                    $valid_value = strtolower($valid_values[$key]);
                    if($value !== $valid_value) {
                        $this->request()->flash('errors', [
                            $key => Lang::get('auth_register_error_' . $key),
                        ]);
                        $this->request()->flash('error', Lang::get('auth_register_error_step1'));
                        return $this->redirect(route('register.index', ['step' => 1]));
                    }
                }

                $_SESSION['register']['step1'] = $inputs->toArray();
                return $this->redirect(route('register.index', ['step' => 2]));
            break;

            // Informations de connexion
            case 2:

                if(!isset($_SESSION['register']['step1'])) {
                    $this->request()->flash('error', Lang::get('auth_register_error_step1'));
                    return $this->redirect(route('register.index', ['step' => 1]));
                }

                $inputs = $this->request()->all()->only([
                    'email',
                    'username',
                    'password',
                    'password_confirmation',
                ]);

                $errors = $this->request()->validate($inputs, [
                    'email' => 'required|email',
                    'username' => 'required|string',
                    'password' => 'required|string',
                    'password_confirmation' => 'required|string',
                ]);

                if(count($errors) > 0) {
                    $this->request()->flash('errors', $errors);
                    return $this->redirect(route('register.index', ['step' => 2]));
                }
                
                $request = $this->request();
                $email = $request->post('email');
                $username = $request->post('username');
                $password = $request->post('password');
                $password_confirmation = $request->post('password_confirmation');
        
                // Check if fields are empty
                if (empty($email) || empty($password) || empty($password_confirmation) || empty($username)) {
                    $request->flash('errors', [
                        'email' => Lang::get('auth_register_error_empty_email'),
                        'username' => Lang::get('auth_register_error_empty_username'),
                        'password' => Lang::get('auth_register_error_empty_password'),
                        'password_confirmation' => Lang::get('auth_register_error_empty_password_confirmation'),
                    ]);
                    return $this->redirect(route('register.index'));
                }
        
                // Check if password match
                if ($password !== $password_confirmation) {
                    $request->flash('errors', [
                        'password' => Lang::get('auth_register_error_password_not_match'),
                        'password_confirmation' => Lang::get('auth_register_error_password_not_match'),
                    ]);
                    return $this->redirect(route('register.index'));
                }
        
                // Check if email already exists
                $user = User::where('email', '=', $email)->first();
                if($user !== null) {
                    $request->flash('errors', [
                        'email' => Lang::get('auth_register_error_email_already_exists'),
                    ]);
                    return $this->redirect(route('register.index'));
                }
        
                // Create user
                $user = new User();
                $user->email = $email;
                $user->username = $username;
                $user->password = password_hash($password, PASSWORD_DEFAULT);
                $user->last_online = date('Y-m-d H:i:s');
                $user->save();
        
                // Login
                Auth::login($user->toArray());
                $request->flash('success', Lang::get('auth_register_success'));

                // Send welcome email
                try {
                    (new WelcomeEmail($user))->send();
                }
                catch(\Exception $e) {
                    Logger::logException($e);
                }
        
                return $this->redirect('/');
            break;
        }

    }
}