<?php

namespace Core;

class App
{
    const VERSION = '1.0.0';

    const BASE_PATH = __DIR__ . '/../../';
    const SRC_PATH = __DIR__ . '/../../src/';
    const CORE_PATH = __DIR__ . '/../../src/Core/';
    const PUBLIC_PATH = __DIR__ . '/../../public/';
    const STORAGE_PATH = __DIR__ . '/../../storage/';
    const CACHE_PATH = __DIR__ . '/../../storage/cache/';
    const LOGS_PATH = __DIR__ . '/../../storage/logs/';
    const UPLOADS_PATH = __DIR__ . '/../../storage/uploads/';


    public function __construct($onlyFirstInit = false)
    {
        try {
            $this->handleErrorsExceptions();
            $this->init($onlyFirstInit);
        } catch (\Throwable $th) {
            throw $th;
        }
    }

    private function init($onlyFirstInit = false)
    {
        // Load alias
        require_once dirname(__FILE__) . '/Alias.php';
        
        // Load env
        Env::load();

        // Set debug mode
        if (Config::get('app.debug', false) === true || Helper::isCLI()) {
            error_reporting(E_ALL);
            ini_set('display_errors', 1);
        } else {
            error_reporting(0);
            ini_set('display_errors', 0);
        }

        // Set timezone
        date_default_timezone_set(Config::get('app.timezone', 'UTC'));

        // Check folders
        $this->ensureNeededFolders();

        // Process models installs and updates
        if (Config::get('database.allow_models_install', false) === true || Config::get('database.allow_models_update', false) === true) {
            // Check if has migration table
            BaseModel::checkMigrationTable();

            // Initialize all models
            $models = glob(dirname(__FILE__) . '/../Models/*.php');
            foreach ($models as $model) {
                $model = explode('/', $model);
                $model = end($model);
                $model = explode('.', $model);
                $model = $model[0];
                $model = '\\Models\\' . $model;
                $model = new $model();
                $model->initialize();
            }
        }

        if ($onlyFirstInit === true)
            return;

        // Load langs
        (Lang::instance())->load();

        if (!Helper::isCLI()) {

            // Load Scheduler if enabled on request
            if (Config::get('scheduler.enabled_on_request', false) === true) {
                (new Scheduler())->run();
            }

            // Load Request
            Request::instance();

            // Check if has flash exception
            if (Request::instance()->hasFlash('exception')) {
                $exception = Request::instance()->flash('exception');
                $this->presentExceptionPage($exception);
                exit;
            }

            // Initialize router
            (new Router())->run();
        } else {

            // Load CLI
            (new \Core\Cli())->run();
        }
    }

    private function ensureNeededFolders()
    {
        // Storage
        $storage = self::STORAGE_PATH;
        if (!file_exists($storage)) {
            mkdir($storage, 0777, true);
        } else {
            if (!is_writable($storage)) {
                if (!chmod($storage, 0777)) {
                    throw new \Exception("Storage folder is not writable");
                }
            }
        }

        // Logs
        $logs = self::LOGS_PATH;
        if (!file_exists($logs)) {
            mkdir($logs, 0777, true);
        } else {
            if (!is_writable($logs)) {
                if (!chmod($logs, 0777)) {
                    throw new \Exception("Logs folder is not writable");
                }
            }
        }

        // Cache
        $cache = self::CACHE_PATH;
        if (!file_exists($cache)) {
            mkdir($cache, 0777, true);
        } else {
            if (!is_writable($cache)) {
                if (!chmod($cache, 0777)) {
                    throw new \Exception("Cache folder is not writable");
                }
            }
        }

        // Uploads
        $uploads = self::UPLOADS_PATH;
        if (!file_exists($uploads)) {
            mkdir($uploads, 0777, true);
        } else {
            if (!is_writable($uploads)) {
                if (!chmod($uploads, 0777)) {
                    throw new \Exception("Uploads folder is not writable");
                }
            }
        }

        // Uploads symbolics links
        $uploadsSymbolicLinks = realpath(self::PUBLIC_PATH) . '/uploads';
        if (!file_exists($uploadsSymbolicLinks)) {
            symlink($uploads, $uploadsSymbolicLinks);
        }
    }

    private function handleErrorsExceptions()
    {
        set_error_handler(function ($errno, $errstr, $errfile, $errline) {
            if (!(error_reporting() & $errno)) {
                // This error code is not included in error_reporting
                return;
            }

            switch ($errno) {
                case E_USER_ERROR:
                    Logger::error("ERROR [$errno] $errstr (in file $errfile on line $errline)");
                    throw new \ErrorException($errstr, $errno, 0, $errfile, $errline);
                    exit(1);

                case E_USER_WARNING:
                    Logger::warning("WARNING [$errno] $errstr (in file $errfile on line $errline)");
                    break;

                case E_USER_NOTICE:
                    Logger::info("NOTICE [$errno] $errstr (in file $errfile on line $errline)");
                    break;

                default:
                    Logger::error("Unknown: [$errno] $errstr (in file $errfile on line $errline)");
                    break;
            }

            // If debug mode
            if (Config::get('app.debug', false) === true) {
                throw new \ErrorException($errstr, $errno, 0, $errfile, $errline);
            }

            /* Don't execute PHP internal error handler */
            return true;
        });

        set_exception_handler(function ($exception) {
            $this->presentExceptionPage($exception);
        });
    }

    private function presentExceptionPage($exception)
    {
        $appName = Config::get('app.name', 'App');
        if (is_string($exception))
            $exception = new \Exception($exception);

        if (Config::get('logger.log_exceptions', true) === true)
            Logger::logException($exception);

        // Render for CLI
        if (Helper::isCLI()) {
            echo 'Whoops, something went wrong' . PHP_EOL;
            echo $exception->getMessage() . PHP_EOL;
            echo $exception->getFile() . ' on line ' . $exception->getLine() . PHP_EOL;
            echo implode('\n#', explode('#', $exception->getTraceAsString()));
            return;
        }

        // HTTP Request - Set 500 status code
        http_response_code(500);

        // Render for AJAX request
        if(Request::instance()->ajax()) {
            if (Config::get('app.debug', false) === true) {
                $exceptionString = '';
                $previous = $exception;
                while ($previous !== null) {
                    $exceptionString .= '<br><br>' . $previous->getMessage() . ' - in file ' . $previous->getFile() . ' - on line ' . $previous->getLine();
                    $exceptionString .= implode('<br>#', explode('#', $previous->getTraceAsString()));
                    $previous = $previous->getPrevious();
                }

                echo json_encode([
                    'success' => false,
                    'error' => true,
                    'message' => $exception->getMessage(),
                    'file' => $exception->getFile(),
                    'line' => $exception->getLine(),
                    'trace' => $exceptionString
                ]);
                return;
            }

            echo json_encode([
                'success' => false,
                'error' => true,
                'message' => 'Whoops, something went wrong'
            ]);
            return;
        }

        // Render for web request
        if (Config::get('app.debug', false) === true) {
            //     $exceptionString = implode('<br>#', explode('#', $exception->getTraceAsString()));
            // Show exception trace with previous
            $exceptionString = '';
            $previous = $exception;
            while ($previous !== null) {
                $exceptionString .= '<br><br>' . $previous->getMessage() . ' - in file ' . $previous->getFile() . ' - on line ' . $previous->getLine();
                $exceptionString .= implode('<br>#', explode('#', $previous->getTraceAsString()));
                $previous = $previous->getPrevious();
            }


            echo <<<HTML
                <html>
                    <head>
                        <title>{$appName} - Fatal error</title>
                        <style>
                            body {
                                background-color: #fff;
                                color: #636b6f;
                                font-family: 'Nunito', sans-serif;
                                font-weight: 100;
                                height: 100vh;
                                margin: 0;
                            }
                            div {
                                text-align: center;
                                padding: 30px;
                                background-color: #f1f1f1;
                                border: 1px solid #ccc;
                                border-radius: 5px;
                                margin: 30px;
                            }
                            h1 {
                                font-size: 36px;
                                margin-bottom: 30px;
                            }
                            h2 {
                                font-size: 24px;
                                margin-bottom: 30px;
                            }
                            p {
                                font-size: 18px;
                                margin-bottom: 30px;
                                display: block;
                                text-align: left;
                                border: 1px solid #ccc;
                                border-radius: 5px;
                                background-color: #fff;
                                padding: 10px;
                                margin: 30px;
                            }
                        </style>
                    </head>
                    <body>
                        <div>
                            <h1>
                                {$appName} - Fatal error
                            </h1>
                            <h2>
                                {$exception->getMessage()} - in file {$exception->getFile()} - on line {$exception->getLine()}
                            </h2>
                        </div>

                        <p>
                            {$exceptionString}
                        </p>
                    </body>
                </html>
            HTML;
            return;
        }

        if (View::exists('errors.500'))
            return View::render('errors.500', ['exception' => $exception]);

        echo <<<HTML
            <html>
                <head>
                    <title>{$appName} - Fatal error</title>
                    <style>
                        body {
                            background-color: #fff;
                            color: #636b6f;
                            font-family: 'Nunito', sans-serif;
                            font-weight: 100;
                            height: 100vh;
                            margin: 0;
                        }
                        h1 {
                            font-size: 36px;
                            margin-bottom: 30px;
                            margin-top: 30px;
                            text-align: center;
                        }
                        p {
                            font-size: 18px;
                            margin-bottom: 30px;
                            display: block;
                            text-align: left;
                            border: 1px solid #ccc;
                            border-radius: 5px;
                            background-color: #fff;
                            padding: 10px;
                            margin: 30px;
                            text-align: center;
                        }
                    </style>
                </head>
                <body>
                    <h1>Whoops</h1>
                    <p>Something went wrong</p>
                </body>
            </html>
        HTML;
    }
}
