<?php
namespace Core;

class BaseModel implements \IteratorAggregate {
    protected $db;

    protected $table = null;
    protected $primaryKey = 'id';
    protected $version = '1.0.0';
    protected $timestamps = true;
    
    private $select = '*';
    private $where = '';
    private $limit = '';
    private $orderBy = '';
    private $groupBy = '';
    private $join = '';
    private $params = [];
    private $touched = [];

    /**
     * @var Collection
     */
    private $datas = null;


    public function __construct(array|int $datas = null) {
        $this->db = Database::getInstance();

        if($this->table === null) {
            $table = explode('\\', get_class($this));
            $table = end($table);
            $table = strtolower($table);
            $this->table = $table . 's';
        }

        $this->datas = new Collection();
        
        if(is_array($datas) && count($datas))
            $this->fill($datas, true, false);

        if(is_int($datas) && $datas > 0) {
            $i = $this->find($datas);
            if($i === null) {
                throw new \Exception('Model ' . get_class($this) . ' with id ' . $datas . ' not found');
            }

            $this->fill((array) $i, false, true);
        }
    }

    public function isLoadedObject() {
        return isset($this->datas->{$this->primaryKey});
    }

    protected function fill(array|Collection $datas = [], $skipOnloaded = false, $skipTouch = false) {

        if(empty($this->datas) || !($this->datas instanceof Collection))
            $this->datas = new Collection();

        foreach($datas as $key => $value) {
            $this->datas->$key = $value;

            if(!$skipTouch)
                $this->touch($key);
        }

        if(!$skipOnloaded) {
            if(isset($this->datas->{$this->primaryKey})) {
                if(method_exists($this, 'onLoaded')) {
                    $this->onLoaded();
                }
            }
        }

        return $this;
    }

    private function reset($withData = true) {
        $this->select = '*';
        $this->where = '';
        $this->limit = '';
        $this->orderBy = '';
        $this->groupBy = '';
        $this->join = '';
        $this->params = [];
        $this->touched = [];

        if($withData === true)
            $this->datas = new Collection();

        return $this;
    }

    private function touch($field) {
        $this->touched[] = $field;

        return $this;
    }


    private function get() {
        $sql = 'SELECT ' . $this->select . ' FROM ' . $this->table . ' ' . $this->join . ($this->where == '' ? '' : ' WHERE ' . $this->where) . ' ' . $this->groupBy . ' ' . $this->orderBy . ' ' . $this->limit;
        $query = $this->db->prepare($sql);
        $query->execute($this->params);
        $results = $query->fetchAll(\PDO::FETCH_OBJ);
        
        foreach($results as $key => $result) {

            $results[$key] = new static();
            $results[$key]->fill((array) $result, false, true);
        }

        return new Collection((array)$results);
    }

    private function find($primaryKeyValue) {
        $this->where($this->primaryKey, $primaryKeyValue);
        return $this->first();
    }

    private function first() {
        $this->limit(1)->orderBy($this->primaryKey);
        $result = $this->get()[0] ?? null;

        return $result;
    }

    private function last() {
        $this->limit(1)->orderBy($this->primaryKey, 'DESC');
        return $this->get()[0] ?? null;
    }

    private function count() {
        $sql = 'SELECT COUNT(*) AS count FROM ' . $this->table . ' ' . $this->join . ($this->where == '' ? '' : ' WHERE ' . $this->where) . ' ' . $this->groupBy . ' ' . $this->orderBy . ' ' . $this->limit;
        $query = $this->db->prepare($sql);
        $query->execute($this->params);
        $result = $query->fetchAll(\PDO::FETCH_OBJ);
        return $result[0]->count ?? 0;
    }
    private function sum($field) {
        $sql = 'SELECT SUM(' . $field . ') AS sum FROM ' . $this->table . ' ' . $this->join . ($this->where == '' ? '' : ' WHERE ' . $this->where) . ' ' . $this->groupBy . ' ' . $this->orderBy . ' ' . $this->limit;
        $query = $this->db->prepare($sql);
        $query->execute($this->params);
        $result = $query->fetchAll(\PDO::FETCH_OBJ);
        return $result[0]->sum ?? 0;
    }

    private function paginate($perPage = 15, $currentPage = null, $options = []) {
        
        $pagination = Pagination::make($perPage, $currentPage, $options);
        $this->limit($pagination->perPage, ($pagination->currentPage - 1) * $pagination->perPage);

        $pagination->setItems($this->get());

        return $pagination;
    }

    private function insert(array|Collection|null $data = null) {

        if($data === null) {
            $data = $this->datas ?? new Collection();
        }else{
            $this->fill($data, true, true);
        }

        if($data instanceof Collection)
            $data = $data->toArray();
        else if(!is_array($data))
            $data = (array) $data;

        // Manage updated_at and created_at
        if($this->timestamps) {
            if(!in_array('updated_at', array_keys($data))) {
                $data['updated_at'] = date('Y-m-d H:i:s');
                $this->touch('updated_at');
            }
            if(!in_array('created_at', array_keys($data))) {
                $data['created_at'] = date('Y-m-d H:i:s');
                $this->touch('created_at');
            }
        }

        $columns = array_keys($data);
        $values = array_values($data);

        $sql = 'INSERT INTO ' . $this->table . ' (' . implode(', ', $columns) . ') VALUES (' . implode(', ', array_fill(0, count($columns), '?')) . ')';
        $query = $this->db->prepare($sql);
        $query->execute($values);

        // Asign the new object to this instance
        $this->reset();
        $this->fill(
            array_merge(
                [
                    $this->primaryKey => $this->db->lastInsertId()
                ],
                $data
            ), false, true
        );

        return $this;
    }

    public function debugTable($table, $var) {
        if($this->table === $table) {
            dd($var);
        }
    }

    private function update(array|Collection|null $data = null, $forceNoTouched = false) {

        if($data === null) {
            $data = $this->datas ?? new Collection();
        }else{
            $this->fill($data, true, true);
        }

        if($data instanceof Collection)
            $data = $data->toArray();
        else if(!is_array($data))
            $data = (array) $data;


        // Manage updated_at column
        if($this->timestamps) {
            $data['updated_at'] = date('Y-m-d H:i:s');
            $this->touch('updated_at');
        }

        // Filters touched fields
        if($this->isLoadedObject() && !$forceNoTouched) {
            $data = array_filter($data, function($key) {
                return in_array($key, $this->touched);
            }, ARRAY_FILTER_USE_KEY);
        }

        // If no data to update, return
        if(empty($data)) {
            $this->reset(false);
            return $this;
        }

        $columns = array_keys($data);
        $values = array_values($data);
        

        $sql = 'UPDATE ' . $this->table . ' SET ';
        foreach($columns as $column) {
            $sql .= $column . ' = ?, ';
        }
        $sql = substr($sql, 0, -2);
        $sql .= ($this->where == '' ? '' : ' WHERE ' . $this->where);
        $query = $this->db->prepare($sql);
        $query->execute(array_merge($values, $this->params));

        // Asign the new object to this instance
        $this->reset(false);
        $this->fill($data, true, true);

        if(!$this->isLoadedObject())
            return null;

        return $this;
    }

    private function delete() {

        // Is deleting instance ?
        if($this->isLoadedObject()) {
            $this->where($this->primaryKey, $this->datas->{$this->primaryKey});

            if(method_exists($this, 'onDelete')) {
                call_user_func([$this, 'onDelete']);
            }
        }

        $sql = 'DELETE FROM ' . $this->table . ($this->where == '' ? '' : ' WHERE ' . $this->where);
        $query = $this->db->prepare($sql);
        $query->execute($this->params);
        return null;
    }

    private function remove() {
        if($this->isLoadedObject())
            return $this->delete();

        throw new \Exception('Can\'t remove a non loaded object');
    }

    private function save() {
        if($this->isLoadedObject()) {
            $this->where($this->primaryKey, $this->datas->{$this->primaryKey});
            return $this->update();
        } else {
            return $this->insert();
        }
    }




    // When call property that not exists in this class
    public function __get($name) {
        
        if(empty($this->datas)) {
            $this->datas = new Collection();
        }

        // Handle dynamic attributes
        if(method_exists($this, 'get'.ucfirst($name).'Attribute')) {
            return call_user_func([$this, 'get'.ucfirst($name).'Attribute'], $this->datas->$name ?? null);
        }

        // Return the value if exists
        if(isset($this->datas->$name)) {
            return $this->datas->$name;
        }

        return null;
    }
    public function __set($name, $value) {
        if(empty($this->datas)) {
            $this->datas = new Collection();
        }

        // Handle dynamic attributes
        if(method_exists($this, 'set'.ucfirst($name).'Attribute')) {
            $value = call_user_func([$this, 'set'.ucfirst($name).'Attribute'], $value);
        }

        $this->datas->$name = $value;
        $this->touch($name);
    }
    
    public function __isset($name) {
        return isset($this->datas->$name);
    }
    public function __unset($name) {
        unset($this->datas->$name);
    }

    public function __call($name, $arguments) {
        return call_user_func_array([$this, $name], $arguments);
    }

    public static function __callStatic($name, $arguments) {
        $model = new static();
        return call_user_func_array([$model, $name], $arguments);
    }

    public function toArray() {
        return $this->datas->toArray();
    }
    public function toJson() {
        return $this->datas->toJson();
    }
    public function dump() {
        return [
            'table' => $this->table,
            'primaryKey' => $this->primaryKey,
            'version' => $this->version,
            'timestamps' => $this->timestamps,
            'attributes' => $this->datas
        ];
    }
    public function __debugInfo() {
        return $this->dump();
    }
    public function __toString() {
        return $this->toJson();
    }
    public function getIterator(): \ArrayIterator {
        return new \ArrayIterator($this->datas);
    }

    private function select($columns) {
        $this->select = $columns;
        return $this;
    }

    private function where($column, $operatorOrValue, $value = null) {
        $operator = $operatorOrValue;
        if($value === null) {
            $value = $operator;
            $operator = '=';
        }

        if($value === null && $operator == '=') {
            $operator = 'IS';
            $value = 'NULL';
        }elseif($value === null && $operator == '!=') {
            $operator = 'IS NOT';
            $value = 'NULL';
        }

        // Where IN operator
        if(is_array($value)) {
            $this->where .= ($this->where == '' ? '' : ' AND ') . $column . ' IN (' . implode(', ', array_fill(0, count($value), '?')) . ')';
            $this->params = array_merge($this->params, $value);
            return $this;
        }

        if($value === "NULL") {
            $this->where .= ($this->where == '' ? '' : ' AND ') . $column . ' ' . $operator . ' NULL';
            return $this;
        }

        // Like operator
        if($operator == 'like') {
            $this->where .= ($this->where == '' ? '' : ' OR ') . $column . ' LIKE ?';
            $this->params[] = $value;
            return $this;
        }

        $this->where .= ($this->where == '' ? '' : ' AND ') . $column . ' ' . $operator . ' ?';
        $this->params[] = $value;
        return $this;
    }

    private function whereDatetime($column, $operatorOrValue, $value = null) {
        $operator = $operatorOrValue;
        if($value === null) {
            $value = $operator;
            $operator = '=';
        }
        $this->where .= ($this->where == '' ? '' : ' AND ') . $column . ' ' . $operator . ' ?';
        $this->params[] = date('Y-m-d H:i:s', $value);
        return $this;
    }

    private function orWhere($column, $operatorOrValue, $value = null) {
        $operator = $operatorOrValue;
        if($value === null) {
            $value = $operator;
            $operator = '=';
        }

        if($value === null && $operator == '=') {
            $operator = 'IS';
            $value = 'NULL';
        }elseif($value === null && $operator == '!=') {
            $operator = 'IS NOT';
            $value = 'NULL';
        }

        // Where IN operator
        if(is_array($value)) {
            $this->where .= ($this->where == '' ? '' : ' OR ') . $column . ' IN (' . implode(', ', array_fill(0, count($value), '?')) . ')';
            $this->params = array_merge($this->params, $value);
            return $this;
        }

        if($value === "NULL") {
            $this->where .= ($this->where == '' ? '' : ' OR ') . $column . ' ' . $operator . ' NULL';
            return $this;
        }

        // Like operator
        if($operator == 'like') {
            $this->where .= ($this->where == '' ? '' : ' OR ') . $column . ' LIKE ?';
            $this->params[] = $value;
            return $this;
        }

        $this->where .= ($this->where == '' ? '' : ' OR ') . $column . ' ' . $operator . ' ?';
        $this->params[] = $value;
        return $this;
    }

    private function orderBy($column, $order = 'ASC') {
        $this->orderBy = 'ORDER BY ' . $column . ' ' . $order;
        return $this;
    }
    private function inRandomOrder() {
        $this->orderBy = 'ORDER BY RAND()';
        return $this;
    }

    private function limit($limit, $offset = null) {
        if($offset === null) {
            $this->limit = 'LIMIT ' . $limit;
        } else {
            $this->limit = 'LIMIT ' . $offset . ', ' . $limit;
        }
        return $this;
    }

    private function take($limit) {
        return $this->limit($limit);
    }

    private function groupBy($column) {
        $this->groupBy = 'GROUP BY ' . $column;
        return $this;
    }

    private function join($table, $first, $operatorOrSecond, $second = null, $type = 'INNER') {
        $operator = $operatorOrSecond;
        if($second === null) {
            $second = $operator;
            $operator = '=';
        }
        $this->join .= $type . ' JOIN ' . $table . ' ON ' . $first . ' ' . $operator . ' ' . $second . ' ';
        return $this;
    }





    // Install
    public function initialize() {
        // Check if need install
        if((Config::get('database.allow_models_install', false) === true || Helper::isCLI()) && !$this->isInstalled()) {
            $timestamps = [
                [
                    'name' => 'created_at',
                    'type' => 'datetime',
                    'null' => false
                ],
                [
                    'name' => 'updated_at',
                    'type' => 'datetime',
                    'null' => false
                ]
            ];

            $this->install([
                [
                    'name' => $this->primaryKey,
                    'type' => 'bigint',
                    'length' => 20,
                    'auto_increment' => true,
                    'null' => false
                ],
                ...($this->timestamps ? $timestamps : [])
            ]);

            Logger::log('Table ' . $this->table . ' installed for model ' . get_class($this) . ' with version ' . $this->version);
        }

        // Check for update
        if(Config::get('database.allow_models_update', false) === false && !Helper::isCLI())
            return;

        $sql = 'SELECT * FROM migrations WHERE migration_name = ?';
        $query = $this->db->prepare($sql);
        $query->execute([$this->migrationName()]);
        $migration = $query->fetch(\PDO::FETCH_OBJ);
        if(!$migration) {
            // Check if method exist
            if(!method_exists($this, 'update_' . str_replace('.', '_', $this->version))) {
                throw new \Exception('Method update_' . str_replace('.', '_', $this->version) . ' not found in model ' . get_class($this));
            }

            call_user_func([$this, 'update_' . str_replace('.', '_', $this->version)]);
            Logger::log('Table ' . $this->table . ' updated for model ' . get_class($this) . ' with version ' . $this->version);
        }
    }
    public static function checkMigrationTable() {
        if(!in_array('migrations', Database::getInstance()->query('SHOW TABLES')->fetchAll(\PDO::FETCH_COLUMN))) {
            $sql = 'CREATE TABLE migrations (';
            $sql .= 'id bigint(20) NOT NULL AUTO_INCREMENT, ';
            $sql .= 'migration_name varchar(255) NOT NULL, ';
            $sql .= 'PRIMARY KEY (id))';
            $query = Database::getInstance()->prepare($sql);
            $query->execute();

            Logger::log('Migration table created');
        }
    }
    public function install($fields) {

        // Loop on fields and add forget attributes
        foreach($fields as $key => $field) {
            $fields[$key] = array_merge([
                'length' => null,
                'null' => true,
                'auto_increment' => false,
                'default' => null
            ], $field);
        }

        // Insert in migration table
        $sql = 'INSERT INTO migrations (migration_name) VALUES (?)';
        $query = $this->db->prepare($sql);
        $query->execute([$this->migrationName()]);

        $sql = 'CREATE TABLE IF NOT EXISTS ' . $this->table . ' (';
        foreach($fields as $field) {
            $sql .= $field['name'] . ' ' . $field['type'];
            if($field['length'] !== null) {
                $sql .= '(' . $field['length'] . ')';
            }
            if($field['null'] === false) {
                $sql .= ' NOT NULL';
            }
            if($field['auto_increment'] === true) {
                $sql .= ' AUTO_INCREMENT';
            }
            if($field['default'] !== null) {
                if(is_string($field['default']))
                    $sql .= ' DEFAULT \'' . $field['default'] . '\'';
                else if(is_bool($field['default']))
                    $sql .= ' DEFAULT ' . ($field['default'] ? '1' : '0');
                else if(is_numeric($field['default']))
                    $sql .= ' DEFAULT ' . $field['default'];
                else if(is_null($field['default']))
                    $sql .= ' DEFAULT NULL';
                else
                    throw new \Exception('Unknown default type for field ' . $field['name'] . ' in model ' . get_class($this) . ' : ' . gettype($field['default']));
            }
            // index
            if(isset($field['index']) && $field['index'] === true) {
                $sql .= ', INDEX ' . $field['name'] . '_index (' . $field['name'] . ')';
            }
            $sql .= ', ';
        }
        $sql .= 'PRIMARY KEY (' . $this->primaryKey . '))';
        $query = $this->db->prepare($sql);
        $query->execute();
    }
    public function updateFields($fields) {

        // // Loop on fields and add forget attributes
        if(!isset($fields['add']))
            $fields['add'] = [];
        if(!isset($fields['edit']))
            $fields['edit'] = [];
        if(!isset($fields['delete']))
            $fields['delete'] = [];

        foreach($fields as $keyType => $fieldsType) {
            foreach($fieldsType as $key => $field) {
                $fields[$keyType][$key] = array_merge([
                    'length' => null,
                    'null' => true,
                    'auto_increment' => false,
                    'default' => null
                ], $field);
            }
        }

        // Update migration table
        $sql = 'INSERT INTO migrations (migration_name) VALUES (?)';
        $query = $this->db->prepare($sql);
        $query->execute([$this->migrationName()]);

        // Add columns
        foreach($fields['add'] as $field) {
            $sql = 'ALTER TABLE ' . $this->table . ' ADD ' . $field['name'] . ' ' . $field['type'];
            if($field['length'] !== null) {
                $sql .= '(' . $field['length'] . ')';
            }
            if($field['null'] === false) {
                $sql .= ' NOT NULL';
            }
            if($field['auto_increment'] === true) {
                $sql .= ' AUTO_INCREMENT';
            }
            if($field['default'] !== null) {
                if(is_string($field['default']))
                    $sql .= ' DEFAULT \'' . $field['default'] . '\'';
                else if(is_bool($field['default']))
                    $sql .= ' DEFAULT ' . ($field['default'] ? '1' : '0');
                else if(is_numeric($field['default']))
                    $sql .= ' DEFAULT ' . $field['default'];
                else if(is_null($field['default']))
                    $sql .= ' DEFAULT NULL';
                else
                    throw new \Exception('Unknown default type for field ' . $field['name'] . ' in model ' . get_class($this) . ' : ' . gettype($field['default']));
            }
            // index
            if(isset($field['index']) && $field['index'] === true) {
                $sql .= ', INDEX ' . $field['name'] . '_index (' . $field['name'] . ')';
            }
            $query = $this->db->prepare($sql);
            $query->execute();
        }

        // Edit columns
        foreach($fields['edit'] as $field) {
            $sql = 'ALTER TABLE ' . $this->table . ' MODIFY ' . $field['name'] . ' ' . $field['type'];
            if($field['length'] !== null) {
                $sql .= '(' . $field['length'] . ')';
            }
            if($field['null'] === false) {
                $sql .= ' NOT NULL';
            }
            if($field['auto_increment'] === true) {
                $sql .= ' AUTO_INCREMENT';
            }
            if($field['default'] !== null) {
                if(is_string($field['default']))
                    $sql .= ' DEFAULT \'' . $field['default'] . '\'';
                else if(is_bool($field['default']))
                    $sql .= ' DEFAULT ' . ($field['default'] ? '1' : '0');
                else if(is_numeric($field['default']))
                    $sql .= ' DEFAULT ' . $field['default'];
                else if(is_null($field['default']))
                    $sql .= ' DEFAULT NULL';
                else
                    throw new \Exception('Unknown default type for field ' . $field['name'] . ' in model ' . get_class($this) . ' : ' . gettype($field['default']));
            }
            // index
            if(isset($field['index']) && $field['index'] === true) {
                $sql .= ', INDEX ' . $field['name'] . '_index (' . $field['name'] . ')';
            }
            $query = $this->db->prepare($sql);
            $query->execute();
        }

        // Delete columns
        foreach($fields['delete'] as $field) {
            $sql = 'ALTER TABLE ' . $this->table . ' DROP COLUMN ' . $field['name'];
            $query = $this->db->prepare($sql);
            $query->execute();
        }
    }
    public function isInstalled() {
        $sql = 'SELECT * FROM migrations WHERE migration_name = ?';
        $query = $this->db->prepare($sql);
        $query->execute([$this->migrationName("1.0.0")]);
        return $query->rowCount() > 0;
    }

    private function migrationName($forceVersion = null) {
        if($forceVersion !== null)
            return $this->table . '_' . $forceVersion;
        return $this->table . '_' . $this->version;
    }
}