<?php

namespace Core;

class Cache
{

    /**
     * Get the cache file path
     * 
     * @param string $key
     * @return string
     */
    private static function getFilePath(string $key): string
    {
        $key = md5($key);
        return App::CACHE_PATH . $key . '.cache';
    }

    /**
     * Check if the cache file exists
     * 
     * @param string $key
     * @return bool
     */
    public static function exists(string $key): bool
    {
        if(!file_exists(self::getFilePath($key))) {
            return false;
        }

        if(!is_readable(self::getFilePath($key))) {
            throw new \Exception("Cache file " . self::getFilePath($key) . " is not readable");
        }

        return true;
    }

    /**
     * Get the cache file content
     * 
     * @param string $key
     * @return mixed
     */
    public static function get(string $key, $default = null)
    {
        if (self::exists($key)) {
            $file = self::getFilePath($key);
            $data = json_decode(file_get_contents($file), true);
            if (!empty($data) && is_array($data) && isset($data['expire'], $data['value']) && ($data['expire'] == 0 || $data['expire'] > time()) && $data['value'] != null) {
                return $data['value'];
            }

            unlink($file);
        }

        return $default;
    }

    /**
     * Set the cache file content
     * 
     * @param string $key
     * @param mixed $value
     * @param int $expire in seconds (0 = never expire)
     * @return void
     */
    public static function set(string $key, $value, int $expire = 0): void
    {
        $file = self::getFilePath($key);
        $data = [
            'expire' => $expire > 0 ? time() + $expire : 0,
            'value' => $value
        ];
        file_put_contents($file, json_encode($data));
    }

    /**
     * Get the cache file content or set it if it doesn't exist
     * 
     * @param string $key
     * @param callable $callback
     * @param int $expire in seconds (0 = never expire)
     * @return mixed
     */
    public static function remember(string $key, callable $callback, int $expire = 0)
    {
        if (self::exists($key)) {
            return self::get($key);
        }

        $value = $callback();
        self::set($key, $value, $expire);
        return $value;
    }

    /**
     * Delete the cache file
     * 
     * @param string $key
     * @return void
     */
    public static function delete(string $key): void
    {
        if (self::exists($key)) {
            unlink(self::getFilePath($key));
        }
    }

    /**
     * Alias of delete(). Delete the cache file
     * 
     * @param string $key
     * @return void
     */
    public static function forget(string $key): void
    {
        self::delete($key);
    }

    /**
     * Delete all cache files
     * 
     * @return void
     */
    public static function deleteAll(): void
    {
        $files = glob(App::CACHE_PATH . '*.cache');
        foreach ($files as $file) {
            unlink($file);
        }
    }

    /**
     * Alias of deleteAll(). Delete all cache files
     * 
     * @return void
     */
    public static function clear(): void
    {
        self::deleteAll();
    }

    /**
     * Delete all expired cache files
     * 
     * @return void
     */
    public static function deleteExpired(): void
    {
        $files = glob(App::CACHE_PATH . '*.cache');
        foreach ($files as $file) {
            $data = json_decode(file_get_contents($file), true);
            if(!isset($data['expire']) || ($data['expire'] != 0 && $data['expire'] < time())) {
                unlink($file);
            }
        }
    }
}
