<?php
namespace Core;

class Collection implements \IteratorAggregate, \ArrayAccess, \Countable, \JsonSerializable, \Stringable
{
    protected $items = [];

    public function __construct(array $items = [])
    {
        $this->items = $items;
    }

    public static function make(array $items = []): Collection
    {
        return new static($items);
    }

    public function getIterator(): \ArrayIterator
    {
        return new \ArrayIterator($this->items);
    }

    public function offsetExists(mixed $offset): bool
    {
        return isset($this->items[$offset]);
    }

    public function offsetGet(mixed $offset): mixed
    {
        return $this->items[$offset] ?? null;
    }

    public function offsetSet(mixed $offset, mixed $value): void
    {
        if (is_null($offset)) {
            $this->items[] = $value;
        } else {
            $this->items[$offset] = $value;
        }
    }

    public function offsetUnset(mixed $offset): void
    {
        unset($this->items[$offset]);
    }

    public function count(): int
    {
        return count($this->items);
    }

    public function __serialize()
    {
        return serialize($this->items);
    }

    public function __unserialize($serialized): void
    {
        $this->items = unserialize($serialized);
    }

    public function jsonSerialize(): array
    {
        return $this->items;
    }

    public function toJson(): string
    {
        return json_encode($this->items);
    }

    public function __toString(): string
    {
        return $this->toJson();
    }

    public function __get(string $key): mixed
    {
        return $this->offsetGet($key);
    }

    public function __set(string $key, mixed $value): void
    {
        $this->offsetSet($key, $value);
    }

    public function __isset(string $key): bool
    {
        return $this->offsetExists($key);
    }

    public function __unset(string $key): void
    {
        $this->offsetUnset($key);
    }

    public function keys(): array
    {
        return array_keys($this->items);
    }

    public function values(): array
    {
        return array_values($this->items);
    }

    public function first(): mixed
    {
        return reset($this->items);
    }

    public function last(): mixed
    {
        return end($this->items);
    }

    public function take(int $limit): Collection
    {
        return new static(array_slice($this->items, 0, $limit));
    }

    public function has(mixed $key): bool
    {
        return $this->offsetExists($key);
    }

    public function isEmpty(): bool
    {
        return empty($this->items);
    }

    public function isNotEmpty(): bool
    {
        return !$this->isEmpty();
    }

    public function merge(array $items): Collection
    {
        return new static(array_merge($this->items, $items));
    }

    public function only(array $keys): Collection
    {
        return new static(array_intersect_key($this->items, array_flip($keys)));
    }

    public function except(array $keys): Collection
    {
        return new static(array_diff_key($this->items, array_flip($keys)));
    }

    public function map(callable $callback): Collection
    {
        return new static(array_map($callback, $this->items));
    }

    public function filter(callable $callback): Collection
    {
        return new static(array_filter($this->items, $callback, ARRAY_FILTER_USE_BOTH));
    }

    public function reduce(callable $callback, mixed $initial = null): mixed
    {
        return array_reduce($this->items, $callback, $initial);
    }

    public function each(callable $callback): Collection
    {
        foreach ($this->items as $key => &$item) {
            $callback($item, $key);
        }
        return $this;
    }

    public function sort(callable $callback): Collection
    {
        $items = $this->items;
        usort($items, $callback);
        return new static($items);
    }

    public function sortBy(string $key): Collection
    {
        $items = $this->items;
        usort($items, function ($a, $b) use ($key) {
            return $a[$key] <=> $b[$key];
        });
        return new static($items);
    }

    public function groupBy(string $key): Collection
    {
        $items = [];
        foreach ($this->items as $item) {
            $items[$item[$key]][] = $item;
        }
        return new static($items);
    }

    public function chunk(int $size): Collection
    {
        return new static(array_chunk($this->items, $size));
    }

    public function pluck(string $key): Collection
    {
        return new static(array_column($this->items, $key));
    }

    public function flip(): Collection
    {
        return new static(array_flip($this->items));
    }

    public function unique(): Collection
    {
        return new static(array_unique($this->items));
    }

    public function shuffle(): Collection
    {
        $items = $this->items;
        shuffle($items);
        return new static($items);
    }

    public function reverse(): Collection
    {
        return new static(array_reverse($this->items));
    }

    public function toArray(): array
    {
        return (array)$this->items;
    }

    public function toObject(): object
    {
        return (object) $this->items;
    }

    public function sum($key = null): int|float
    {
        if (is_null($key)) {
            return array_sum($this->items);
        }
        return array_sum($this->pluck($key)->toArray());
    }

    public function avg($key = null): int|float
    {
        if (is_null($key)) {
            return array_sum($this->items) / count($this->items);
        }
        return array_sum($this->pluck($key)->toArray()) / count($this->items);
    }

    public function min($key = null): int|float
    {
        if (is_null($key)) {
            return min($this->items);
        }
        return min($this->pluck($key)->toArray());
    }

    public function max($key = null): int|float
    {
        if (is_null($key)) {
            return max($this->items);
        }
        return max($this->pluck($key)->toArray());
    }

    public function median($key = null): int|float
    {
        $count = count($this->items);
        if ($count === 0) {
            return 0;
        }
        $middle = floor(($count - 1) / 2);
        if ($count % 2) {
            return $this->sortBy($key)[$middle];
        }
        return ($this->sortBy($key)[$middle] + $this->sortBy($key)[$middle + 1]) / 2;
    }

    /**
     * Returns the most common value in the array.
     *
     * @param string $key
     * @return integer|float
     */
    public function mode($key = null): int|float
    {
        $count = count($this->items);
        if ($count === 0) {
            return 0;
        }
        $values = array_count_values($this->sortBy($key)->toArray());
        arsort($values);
        return key($values);
    }

    public function diff(array $items): Collection
    {
        return new static(array_diff($this->items, $items));
    }

    public function intersect(array $items): Collection
    {
        return new static(array_intersect($this->items, $items));
    }

    public function search(mixed $value, bool $strict = false): mixed
    {
        return array_search($value, $this->items, $strict);
    }

    public function searchKey(mixed $value, bool $strict = false): mixed
    {
        return array_search($value, array_keys($this->items), $strict);
    }

    public function searchRecursive(mixed $value, bool $strict = false): mixed
    {
        foreach (new \RecursiveIteratorIterator(new \RecursiveArrayIterator($this->items)) as $key => $item) {
            if ($item === $value && (!$strict || $item === $value)) {
                return $key;
            }
        }
        return false;
    }
}