<?php
namespace Core;

class Helper {

    static $timers = [];
    public static function resetTimers() {
        self::$timers = [];
    }
    public static function resetTimer($name) {
        if(isset(self::$timers[$name])) {
            unset(self::$timers[$name]);
        }
    }
    public static function startTimer($name) {
        self::$timers[$name]['start'] = microtime(true);
    }
    public static function stopTimer($name, $show = false) {
        self::$timers[$name]['end'] = microtime(true);

        if($show) {
            self::showTimer($name);
        }
    }
    public static function getTimer($name) {
        if(!isset(self::$timers[$name]['end'])) {
            self::stopTimer($name);
        }
        
        $time = self::$timers[$name]['end'] - self::$timers[$name]['start'];
        Logger::debug("[TIMER/{$name}] Took $time s");

        return $time;
    }
    public static function showTimer($name) {

        if(!isset(self::$timers[$name]['end'])) {
            self::stopTimer($name);
        }

        $time = self::getTimer($name);

        echo "<div style='background-color: #f5f5f5; border-radius: 5px; padding: 5px; margin-bottom: 10px;'>";
        echo "<code>";
        echo "<span style='color: #a50000; font-weight: bolder;'>$name</span> : ";
        echo "<span style='color: #5825b6; font-weight: lighter;'>$time</span>";
        echo "</code>";
        echo "</div>";
    }

    /**
     * Dump variables
     *
     * @param mixed $vars
     * @return void
     */
    public static function dump($vars) {
        self::prettyDump($vars);
    }

    /**
     * Dump variables and exit
     *
     * @param mixed $vars
     * @return void
     */
    public static function dd($vars) {
        self::dump($vars);

        if(Helper::isCLI())
            return;
            
        exit();
    }

    public static function prettyDump($var) {
        ini_set("highlight.keyword", "#a50000;  font-weight: bolder");
        ini_set("highlight.string", "#5825b6; font-weight: lighter; ");

        ob_start();

        echo "<style>
            .collection {
                background-color: #f5f5f5;
                border-radius: 5px;
                margin-bottom: 10px;
            }
            .collection > code {
                display: block;
                padding: 5px;
                margin: 0;
                border-radius: 5px;
                background-color: #fff;
                border: 1px solid #ccc;
                font-size: 14px;
                font-family: monospace;
                line-height: 1.5;
                overflow-x: auto;
                text-align: left;
            }
            .collection > code .span {
                
            }
        </style>";
        echo "<div class='collection'>";
            highlight_string("<?php\n" . var_export($var, true) . "?>");
        echo "</div>";

        $highlighted_output = ob_get_clean();
        $highlighted_output = str_replace( ["&lt;?php","?&gt;", '::', '__set_state'] , '', $highlighted_output );
        echo $highlighted_output;
    }

    /**
     * Slugify a string
     * 
     * @param string $str String to slugify
     * @param string $delimiter Delimiter
     * @return string
     */
    public static function slugify($str, $delimiter = '-') {
        $str = mb_strtolower(trim($str));
        $str = preg_replace('/[^a-z0-9-]/', $delimiter, $str);
        $str = preg_replace('/-+/', $delimiter, $str);
        return $str;
    }

    /**
     * Generate a random string
     * 
     * @param int $length Length of the string
     * @return string
     */
    public static function randomString($length = 10) {
        $characters = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
        $charactersLength = strlen($characters);
        $randomString = '';
        
        for($i = 0; $i < $length; $i++) {
            $randomString .= $characters[rand(0, $charactersLength - 1)];
        }
        
        return $randomString;
    }

    /**
     * Generate a unique ID
     * 
     * @return string
     */
    public static function uniqueID() {
        return md5(uniqid(rand(), true));
    }

    /**
     * Check if the current request is a CLI request
     * 
     * @return boolean
     */
    public static function isCLI() {
        return (php_sapi_name() === 'cli');
    }

    /**
     * Format a date with a specific format
     * 
     * @param string $date Date to format
     * @param string $format Date format (default: Y-m-d H:i:s) or auto to use the user's language format
     * @return string
     */
    public static function dateFormat($date, $format = 'Y-m-d H:i:s') {

        if(strtolower($format) === 'auto') {
            switch(Lang::instance()->getLang()) {
                case 'fr':
                    $format = 'd/m/Y H:i:s';
                    break;
                default:
                    $format = 'Y-m-d H:i:s';
                    break;
            }
        }

        return date($format, strtotime($date));
    }

    public static function timeDuration($seconds) {
        $dtF = new \DateTime('@0');
        $dtT = new \DateTime("@$seconds");
        return $dtF->diff($dtT)->format('%a:%h:%i:%s');
    }

    /**
     * Redirect to a specific URL
     * 
     * @param string $url URL to redirect to
     * @return void
     */
    public static function redirect($url) {
        header('Location: ' . $url);
        exit();
    }

    /**
     * Get human readable file size
     * 
     * @param int $bytes Size in bytes
     * @param int $decimals Number of decimals
     * 
     * @return string
     */
    public static function humanSize($bytes, $decimals = 2) {
        $sz = 'BKMGTP';
        $factor = floor((strlen($bytes) - 1) / 3);
        return sprintf("%.{$decimals}f", $bytes / pow(1024, $factor)) . @$sz[$factor];
    }

    /**
     * A function that return number in (T, M, K...)
     *
     * @param string|int $n
     * @return string
     */
    public static function numberFormat($n) {

        if(!$n) {
            return 0;
        }

        if($n < 1000) {
            return round($n, 2);
        }else if($n < 999999) {
            return round($n / 1000, 1) . 'K';
        }else if($n < 999999999) {
            return round($n / 1000000, 1) . 'M';
        }else if($n < 999999999999) {
            return round($n / 1000000000, 1) . 'B';
        }else if($n < 999999999999999) {
            return round($n / 1000000000000, 1) . 'T';
        }

        return $n;
    }

    public static function minMax($n, $min, $max) {
        return max($min, min($n, $max));
    }
}