<?php
namespace Core;

class Request {
    public static $singleton = null;
    public static $oldInputs = [];

    public function __construct() {
        self::$singleton = $this;

        if (isset($_SESSION['old_inputs'])) {
            self::$oldInputs = new Collection($_SESSION['old_inputs']);
            unset($_SESSION['old_inputs']);
        }

        $_SESSION['old_inputs'] = $this->request()->toArray();
    }

    public static function instance() {
        if (self::$singleton === null) {
            self::$singleton = new Request();
        }
        return self::$singleton;
    }

    /**
     * Get all request data
     *
     * @return Collection
     */
    public function request() {
        return new Collection($_REQUEST);
    }

    /**
     * Get all request data
     *
     * @return Collection
     */
    public function all() {
        return $this->request();
    }

    /**
     * Get a request value from $_GET
     *
     * @param string|null $key
     * @return Collection|mixed|null
     */
    public function get($key = null) {

        if($key === null) {
            return new Collection($_GET);
        }

        return $_GET[$key] ?? null;
    }

    /**
     * Get a request value from $_POST
     *
     * @param string|null $key
     * @return Collection|mixed|null
     */
    public function post($key = null) {

        if($key === null) {
            return new Collection($_POST);
        }

        return $_POST[$key] ?? null;
    }

    /**
     * Get a request value from $_FILES
     *
     * @param string|null $key
     * @return mixed|null
     */
    public function file($key = null) {

        if($key === null) {
            return $_FILES;
        }

        return $_FILES[$key] ?? null;
    }

    /**
     * Store the given file in the uploads folder and return the path
     * 
     * @param array|string $file File to store or file key
     * @param string $path Path to store the file in the uploads folder
     * @return string|boolean
     */
    public function storeFile($file, $path = "") {
        if(!is_array($file)) {
            $file = $this->file($file);
        }
        
        if($file['error'] !== 0) {
            return false;
        }

        $path = App::UPLOADS_PATH . $path;
        if(!is_dir($path)) {
            mkdir($path, 0777, true);
        }

        $filename = uniqid() . '.' . pathinfo($file['name'], PATHINFO_EXTENSION);
        if(!move_uploaded_file($file['tmp_name'], $path . $filename)) {
            return false;
        }

        return $path . $filename;
    }

    /**
     * Abort the request with a code and a message
     * 
     * @param int $code HTTP code
     * @param string $message Message to display
     * @return void
     */
    public function abort($code = 404, $message = null) {
        http_response_code($code);

        if(View::exists('errors.' . $code))
            View::render('errors.' . $code, ['message' => $message]);
        else echo $message;

        exit;
    }
    
    /**
     * Check if the request is an ajax request
     * 
     * @return boolean
     */
    public function ajax() {
        return !empty($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) == 'xmlhttprequest';
    }

    /**
     * Set a cookie
     * 
     * @param string $name Cookie name
     * @param string $value Cookie value
     * @param int $expire Cookie expire time
     * @param string $path Cookie path
     * @param string $domain Cookie domain
     * @param boolean $secure Cookie secure
     * @param boolean $httponly Cookie httponly
     * @return void
     */
    public function setCookie($name, $value, $expire = 0, $path = '/', $domain = "", $secure = false, $httponly = true) {
        setcookie($name, $value, $expire, $path, $domain, $secure, $httponly);
    }

    /**
     * Get a cookie
     * 
     * @param string $key Cookie key
     * @return mixed|null
     */
    public function cookie($key) {
        return $_COOKIE[$key] ?? null;
    }

    /**
     * Check if a cookie exists
     * 
     * @param string $key Cookie key
     * @return boolean
     */
    public function hasCookie($key) {
        return isset($_COOKIE[$key]);
    }

    /**
     * Delete a cookie
     * 
     * @param string $key Cookie key
     * @return void
     */
    public function deleteCookie($key) {
        unset($_COOKIE[$key]);
        setcookie($key, null, -1);
    }

    /**
     * Get or Set a session value. Without parameters, return all session values
     * 
     * @param string|null $key Session key
     * @param mixed|null $value Session value
     * @return Collection|mixed|null
     */
    public function session($key = null, $value = null) {

        if($value !== null && $key !== null) {
            $_SESSION[$key] = $value;
        }

        if($key !== null) {
            return $_SESSION[$key] ?? null;
        }
        
        return new Collection($_SESSION);
    }

    /**
     * Return the current auth user
     * 
     * @return \Models\User|null
     */
    public function user() {
        return Auth::user();
    }

    /**
     * Return the current auth class
     * 
     * @return \Core\Auth
     */
    public function auth() {
        return Auth::class;
    }

    /**
     * Set or Get a flash message from session
     * 
     * @param string $key Flash key
     * @param mixed|null $value Flash value
     */
    public function flash($key, $value = null) {

        $keys = explode('.', $key);
        $key = null;
        $session = &$_SESSION;

        foreach ($keys as $k) {
            $key = $k;
            if(!isset($session[$key])) {
                $session[$key] = [];
            }
            $session = &$session[$key];
        }

        if(empty($value)) {
            $value = $session;
            $session = null;
            unset($session);

            $toEval = array_map(function($k) {
                return "['$k']";
            }, $keys);

            $toEval = implode('', $toEval);
            eval("unset(\$_SESSION$toEval);");

            return $value;
        }

        $session = $value;

        // if ($value === null) {
        //     $value = $_SESSION[$key] ?? null;

        //     if($value !== null)
        //         unset($_SESSION[$key]);

        //     return $value;
        // } else {
        //     $_SESSION[$key] = $value;
        // }
    }

    /**
     * Show a flash message from session if exists and delete it
     * 
     * @param string $key Flash key
     * @return mixed|null
     */
    public static function showFlash($key) {
        return self::instance()->flash($key);
    }

    /**
     * Check if a flash message exists
     * 
     * @param string $key Flash key
     * @return boolean
     */
    public function hasFlash($key) {
        return isset($_SESSION[$key]);
    }

    /**
     * Get an old input value
     * 
     * @param string $key Input key
     * @return mixed|null
     */
    public static function old($key) {
        return self::$oldInputs[$key] ?? null;
    }

    /**
     * Validate a request
     * 
     * @param array $data Data to validate
     * @param array $rules Rules to apply
     * @return array
     */
    public static function validate(array|Collection $data, $rules) {
        $errors = [];
        foreach ($rules as $key => $rule) {
            $value = $data[$key] ?? null;
            $rules = explode('|', $rule);
            foreach ($rules as $rule) {
                $rule = explode(':', $rule);
                $ruleName = $rule[0];
                $ruleValue = $rule[1] ?? null;
                $error = null;
                switch ($ruleName) {
                    case 'required':
                        if ($value === null || $value === '') {
                            $error = Lang::get('validation.required', ['attribute' => $key]);
                        }
                        break;
                    case 'min':
                        if (empty($value) || strlen($value) < $ruleValue) {
                            $error = Lang::get('validation.min', ['attribute' => $key, 'min' => $ruleValue]);
                        }
                        break;
                    case 'max':
                        if (strlen($value) > $ruleValue) {
                            $error = Lang::get('validation.max', ['attribute' => $key, 'max' => $ruleValue]);
                        }
                        break;
                    case 'email':
                        if (!filter_var($value, FILTER_VALIDATE_EMAIL)) {
                            $error = Lang::get('validation.email', ['attribute' => $key]);
                        }
                        break;
                    case 'unique':
                        $model = new $ruleValue;
                        $result = $model->where($key, $value)->first();
                        if ($result) {
                            $error = Lang::get('validation.unique', ['attribute' => $key]);
                        }
                        break;
                    case 'exists':
                        $model = new $ruleValue;
                        $result = $model->where($key, $value)->first();
                        if (!$result) {
                            $error = Lang::get('validation.exists', ['attribute' => $key]);
                        }
                        break;
                    case 'confirmed':
                        if ($value !== $data[$key . '_confirmation']) {
                            $error = Lang::get('validation.confirmed', ['attribute' => $key]);
                        }
                        break;
                    case 'image':
                        if (!preg_match('/^data:image\/(\w+);base64,/', $value)) {
                            $error = Lang::get('validation.image', ['attribute' => $key]);
                        }
                        break;
                    case 'in':
                        $values = explode(',', $ruleValue);
                        if (!in_array($value, $values)) {
                            $error = Lang::get('validation.in', ['attribute' => $key]);
                        }
                        break;
                    case 'integer':
                        if (!is_numeric($value)) {
                            $error = Lang::get('validation.integer', ['attribute' => $key]);
                        }
                        break;
                    case 'string':
                        if (empty($value) || !is_string($value)) {
                            $error = Lang::get('validation.string', ['attribute' => $key]);
                        }
                        break;
                    case 'date':
                        if (!strtotime($value)) {
                            $error = Lang::get('validation.date', ['attribute' => $key]);
                        }
                        break;
                    default:
                        break;
                }

                if ($error) {
                    $errors[$key] = $error;
                }
            }
        }
        
        Request::instance()->flash('errors_input', $errors);

        return $errors;
    }
}