<?php
namespace Core;

use Core\Exceptions\PageNotFoundException;

class Router {
    public static function __callStatic($name, $arguments) {
        return call_user_func_array([Core_Router::getInstance(), $name], $arguments);
    }

    public static function getInstance() {
        return Core_Router::getInstance();
    }

    public function __call($name, $arguments) {
        return call_user_func_array([Core_Router::getInstance(), $name], $arguments);
    }

    public function __get($name) {
        return Core_Router::getInstance()->$name;
    }

    public function __set($name, $value) {
        Core_Router::getInstance()->$name = $value;
    }

    public function __isset($name) {
        return isset(Core_Router::getInstance()->$name);
    }

    public function __unset($name) {
        unset(Core_Router::getInstance()->$name);
    }

    public function __construct() {
        Core_Router::getInstance();
    }

    public static function run() {
        Core_Router::getInstance()->run();
    }

    /**
     * Get a route URL by name
     *
     * @param string $name Route name
     * @param array $params Params to set on route
     * @return string|null
     */
    public static function route($name, $params = []) {
        return Core_Router::getInstance()->route($name, $params);
    }

    /**
     * Add a prefix to the next routes added
     * 
     * @param string $prefix
     * @param array|string $middlewares
     * @return void
     **/
    public static function prefix($prefix, $middlewares = []) {
        Core_Router::getInstance()->prefix($prefix, $middlewares);
    }

    /**
     * End the current prefix
     * 
     * @return void
     **/
    public static function endPrefix() {
        Core_Router::getInstance()->endPrefix();
    }

    /**
     * Get current route data
     * 
     * @return array
     **/
    public static function current() {
        return Core_Router::getInstance()->current();
    }

    /**
     * Check if current route is the given name
     * 
     * @param string $name
     * @return bool
     **/
    public static function isName(string $name) {
        return Core_Router::getInstance()->isName($name);
    }

    /**
     * Check if current route is the given uri
     * 
     * @param string $uri
     * @return bool
     **/
    public static function isUri(string $uri) {
        return Core_Router::getInstance()->isUri($uri);
    }

    /**
     * Check if current route is the given uri
     * 
     * @param string $uri
     * @return bool
     **/
    public static function isRoute(string $uri) {
        return Core_Router::getInstance()->isUri($uri);
    }

    /**
     * Get current route data
     * 
     * @return array
     **/
    public static function asset($path) {
        return Config::get('app.url', '') . '/assets/' . ltrim($path, '/');
    }
}

class Core_Router {
    public static $singleton;
    private $routes = [];
    private $prefix = [];
    private $middlewares = [];
    private $current;

    public static function getInstance() {
        if(self::$singleton === null)
            self::$singleton = new self();

        return self::$singleton;
    }

    public function __construct() {
        self::$singleton = $this;
        $this->init();
    }

    private function init() {
        $this->initRoutes();
    }

    private function initRoutes() {

        // Cached ?
        if(Config::get('router.cache_enabled', false) === true) {
            $this->routes = Cache::get('routes');
            if($this->routes !== null) {
                return;
            }
        }

        $this->routes = [];
        require_once dirname(__FILE__) . '/../routes.php';

        // Save cache
        if(Config::get('router.cache_enabled', false) === true) {
            Cache::set('routes', $this->routes, Config::get('router.cache_duration', 0));
        }
    }

    public function add($method, $uri, $controller, $action = null, $name = null, $middlewares = []) {
        $uri = $this->checkPrefix($uri);
        $middlewares = $this->checkMiddlewares($middlewares);

        $tmp = explode('@', $controller);
        if(count($tmp) == 2) {
            $name = $action;
            $middlewares = is_array($name) ? $name : [];

            $controller = $tmp[0] ?? $controller;
            $action = $tmp[1] ?? null;
        }

        if($action === null) {
            throw new \Exception("Route action is null for route $uri");
        }

        $this->routes[] = [
            'uri' => $uri,
            'method' => $method,
            'controller' => $controller,
            'action' => $action,
            'name' => $name ?? $controller . '@' . $action,
            'middlewares' => $middlewares,
        ];
    }
    public function get($uri, $controller, $action = null, $name = null, $middlewares = []) {
        $this->add('GET', $uri, $controller, $action, $name, $middlewares);
    }
    public function post($uri, $controller, $action = null, $name = null, $middlewares = []) {
        $this->add('POST', $uri, $controller, $action, $name, $middlewares);
    }
    public function delete($uri, $controller, $action = null, $name = null, $middlewares = []) {
        $this->add('DELETE', $uri, $controller, $action, $name, $middlewares);
    }
    public function put($uri, $controller, $action = null, $name = null, $middlewares = []) {
        $this->add('PUT', $uri, $controller, $action, $name, $middlewares);
    }
    public function resource($uri, $controller) {
        $this->get($uri, $controller, 'index');
        $this->get($uri . '/create', $controller, 'create');
        $this->post($uri, $controller, 'store');
        $this->get($uri . '/{id}', $controller, 'show');
        $this->get($uri . '/{id}/edit', $controller, 'edit');
        $this->put($uri . '/{id}', $controller, 'update');
        $this->delete($uri . '/{id}', $controller, 'destroy');
    }

    private function checkPrefix($uri) {

        $global_prefix = "";
        if(Config::get('router.global_prefix', null) !== null) {
            $global_prefix = Config::get('router.global_prefix');
        }
        
        if(count($this->prefix) > 0) {
            $uri = implode('/', $this->prefix) . '/' . ltrim($uri, '/');
        }

        if(!empty($global_prefix)) {
            $uri = $global_prefix . '/' . ltrim($uri, '/');
        }

        // Remove double slashes and trim end slash
        $uri = preg_replace('/\/+/', '/', $uri);
        $uri = rtrim($uri, '/');

        if(empty($uri))
            $uri = '/';

        return $uri;
    }

    private function checkMiddlewares($middlewares) {
        if(!is_array($middlewares))
            $middlewares = [$middlewares];

        foreach($this->middlewares as $middleware) {
            $middlewares = array_merge($middlewares, $middleware);
        }

        return $middlewares;
    }

    public function route($name, $params = []) {

        foreach($this->routes as $route) {
            if($route['name'] == $name) {
                $uri = $route['uri'];
                $params_left = $params;
                foreach($params as $key => $value) {
                    $uri = str_replace('{' . $key . '}', $value, $uri, $count);
                    
                    if($count > 0)
                        unset($params_left[$key]);
                }

                if(count($params_left) > 0) {
                    $uri .= '?' . http_build_query($params_left);
                }

                $route = trim(Config::get('app.url', '') . '/' . $uri, '/');
                // Remove extra slashes (not http:// or https://)
                $route = preg_replace('/([^:])(\/{2,})/', '$1/', $route);

                if(empty($route))
                    $route = '/';

                return $route;
            }
        }

        throw new \Exception("Route URL $name not found");
        return null;
    }

    /**
     * Get current route data
     * 
     * @return array
     **/
    public function current() {
        return $this->current;
    }

    /**
     * Check if current route is the given name
     * 
     * @param string $name
     * @return bool
     **/
    public function isName(string $name) {
        return $this->current['name'] == $name;
    }

    /**
     * Check if current route is the given uri
     * 
     * @param string $uri
     * @return bool
     **/
    public function isUri(string $uri) {
        return $this->current['uri'] == $uri;
    }

    public function prefix(string $prefix, string|array $middlewares = []) {
        $this->prefix[] = $prefix;

        if(!is_array($middlewares))
            $middlewares = [$middlewares];

        $this->middlewares[] = $middlewares;
    }

    public function endPrefix() {
        // Remove last element
        array_pop($this->prefix);
        array_pop($this->middlewares);
    }

    public function initController($controller) {

        if(is_callable($controller))
            return $controller;
        
        if(is_string($controller)) {
            $controller = 'Controllers\\' . $controller;

            try {
                if(class_exists($controller)) {

                    $middlewares = $this->current['middlewares'];
                    if(count($middlewares)) {
                        foreach($middlewares as $middleware) {
                            Middleware::add($middleware);
                        }
                    }

                    $c = new $controller();
                    $c->init();

                    return $c;
                }
            }
            catch(\Exception $e) {
                throw new \Exception("Controller not found: $controller");
            }
        }

        throw new \Exception("Controller not found: $controller");        
    }

    public function run() {
        $uri = $_SERVER['REQUEST_URI'];
        $method = $_SERVER['REQUEST_METHOD'];


        if($global_prefix = Config::get('router.global_prefix', null)) {
            // If current uri doesn't start with global prefix, redirect
            if($uri === "/") {
                // Redirect to global prefix
                header('Location: ' . $global_prefix);
                exit;
            }
        }


        // is delete ?
        if($method == 'POST' && isset($_POST['_method']) && $_POST['_method'] == 'DELETE') {
            $method = 'DELETE';
        }
        // is put ?
        else if($method == 'POST' && isset($_POST['_method']) && $_POST['_method'] == 'PUT') {
            $method = 'PUT';
        }


        // Clean search params
        $uri = explode('?', $uri)[0];

        $route = $this->findRoute($uri, $method);
        if($route) {
            $this->current = array_merge($route, [
                'raw' => [
                    'method' => $method,
                    'path' => $uri,
                    'hostname' => $_SERVER['HTTP_HOST'],
                    'query_string' => $_SERVER['QUERY_STRING'] ?? '',
                    'uri' => $_SERVER['REQUEST_URI'],
                    'url' => Config::get('app.url', '') . $_SERVER['REQUEST_URI'],
                ],
            ]);

            $controller = $this->initController($route['controller']);
            $action = $route['action'];
            $params = isset($route['params']) && is_array($route['params']) ? $route['params'] : [];

            Logger::debug("Router: [{$method}] {$uri} -> {$route['controller']}@{$route['action']} - from: " . $_SERVER['REMOTE_ADDR'] . " - path_params: " . json_encode($params) . " - query_params: " . json_encode($_GET) . " - body_params: " . json_encode($_POST));
            if($action === null && is_callable($controller)) {
                call_user_func_array($controller, $params);
                return;
            }

            if(!method_exists($controller, $action)) {
                if($this->error(404) !== false)
                    return;

                http_response_code(404);
                throw new \Exception("Undefined method: $action in " . get_class($controller));
            }

            call_user_func_array([$controller, $action], $params);
            
        } else {
            if($this->error(404) !== false)
                return;
                
            http_response_code(404);
            throw new PageNotFoundException("Undefined route: $uri");
        }
    }

    private function findRoute($uri, $method) {
        foreach($this->routes as $route) {
            if($route['method'] != $method)
                continue;

            // Check if uri correspond
            // A route name can be with eg: /path/{key}/view with dynamic key
            // So we need to check if the uri match with the route
            $routeUri = $route['uri'];
            $routeUri = preg_replace('/{[a-zA-Z0-9\-]+}/', '([a-zA-Z0-9\-]+)', $routeUri);
            $routeUri = str_replace('/', '\/', $routeUri);
            $routeUri = '/^' . $routeUri . '$/';
            
            if(preg_match($routeUri, $uri, $matches)) {
                // Remove first element
                array_shift($matches);

                // Get params
                $params = [];
                foreach($matches as $key => $match) {
                    $params[$key] = $match;
                }

                // Set params on method
                if($method == 'GET')
                    $_GET = array_merge($_GET, $params);
                else if($method == 'POST' || $method == 'PUT' || $method == 'DELETE')
                    $_POST = array_merge($_POST, $params);

                return array_merge($route, ['params' => $params]);
            }
        }

        return false;
    }

    private function error($code) {
        http_response_code($code);
        if(View::exists('errors.' . $code)) {
            return View::render('errors.' . $code);
        }

        return false;
    }
}